from archicat import components

from enum import EnumType,Enum
from typing import Optional,TYPE_CHECKING,Any,Callable,Type,Self

if TYPE_CHECKING:
    from .transformer import ScratchFileBuilder

class BlockAccessModifier(Enum):
    ALL = 'all'
    STAGE = 'stage'
    SPRITE = 'sprite'

def _enum_has_value(enum: EnumType,value: Any) -> bool:
    return value in set(item.value for item in enum)


type Value = str | float | int

type InputItem = components.Id | Value | None

type FieldOrInput = Callable[['ScratchFileBuilder',InputItem],components.Input | components.Field]


def field(builder: 'ScratchFileBuilder',value: str,id: Optional[components.Id] = None) -> components.Field:
    return components.Field(value,id)

def variable_field(builder: 'ScratchFileBuilder',name: str) -> components.Field:
    return field(builder,name,builder._get_variable(name))

def list_field(builder: 'ScratchFileBuilder',name: str) -> components.Field:
    return field(builder,name,builder._get_list(name))

def message_field(builder: 'ScratchFileBuilder',name: str) -> components.Field:
    return field(builder,name,builder._get_message(name))

def costume_field(builder: 'ScratchFileBuilder',name: str) -> components.Field:
    return field(builder,name)

def sound_field(builder: 'ScratchFileBuilder',name: str) -> components.Field:
    return field(builder,name)

def option_field(options: Optional[Type[EnumType]] = None) -> FieldOrInput:
    def _option_field(builder: 'ScratchFileBuilder',value: str | Enum) -> components.Field:
        if isinstance(value,options):
            return field(builder,value.value)
        elif isinstance(value,str) and (options is None or _enum_has_value(options,value)):
            return field(builder,value)
    return _option_field


def input(builder: 'ScratchFileBuilder',type: components.InputType | int,
          value: InputItem,shadow: Optional[InputItem] = None) -> components.Input:
    if isinstance(value,components.Id):
        if isinstance(block := builder._block_by_id(value),components.Block):
            block.parent = builder.parent_block_stack[-1]
        if components.InputType(type) == components.InputType.SHADOW:
            builder._block_by_id(value).shadow = True
    if isinstance(shadow,components.Id):
        builder._block_by_id(value).parent = builder.parent_block_stack[-1]
        builder._block_by_id(value).shadow = True
    return components.Input(type,value,shadow)

def bool_input(builder: 'ScratchFileBuilder',value: components.Id) -> components.Input:
    return input(builder,components.InputType.NO_SHADOW,value)

def _value_input(builder: 'ScratchFileBuilder',type: components.InputType,value: Any) -> components.Input:
    if isinstance(value,components.Id):
        return input(builder,components.InputType.OBSCURED_SHADOW,value,
                     components.Value(type,''))
    else:
        return input(builder,components.InputType.SHADOW,
                     components.Value(type,value))
    
def int_input(builder: 'ScratchFileBuilder',value: int | components.Id) -> components.Input:
    return _value_input(builder,components.ValueType.INT,value)

def float_input(builder: 'ScratchFileBuilder',value: float | components.Id) -> components.Input:
    return _value_input(builder,components.ValueType.FLOAT,value)

def unsigned_int_input(builder: 'ScratchFileBuilder',value: int | components.Id) -> components.Input:
    return _value_input(builder,components.ValueType.UNSIGNED_INT,value)

def unsigned_float_input(builder: 'ScratchFileBuilder',value: float | components.Id) -> components.Input:
    return _value_input(builder,components.ValueType.UNSIGNED_FLOAT,value)

def angle_input(builder: 'ScratchFileBuilder',value: float | components.Id) -> components.Input:
    return _value_input(builder,components.ValueType.ANGLE,value)

def color_input(builder: 'ScratchFileBuilder',value: str | components.Id) -> components.Input:
    return _value_input(builder,components.ValueType.COLOR,value)

def string_input(builder: 'ScratchFileBuilder',value: str | components.Id) -> components.Input:
    return _value_input(builder,components.ValueType.STRING,value)

def chain_input(builder: 'ScratchFileBuilder',value: components.Id) -> components.Input:
    return input(builder,components.InputType.NO_SHADOW,value)

def costume_input(builder: 'ScratchFileBuilder',value: components.Id | str) -> components.Input:
    input_block = Block('looks_costume',COSTUME=field)
    if isinstance(value,components.Id):
        return input(builder,components.InputType.OBSCURED_SHADOW,value,
                     input_block(builder,builder.current_target.costumes[0].name))
    else:
        return input(builder,components.InputType.SHADOW,input_block(builder,value))
    
def sound_input(builder: 'ScratchFileBuilder',value: components.Id | str) -> components.Input:
    input_block = Block('sound_sounds_menu',SOUNDS_MENU=field)
    if isinstance(value,components.Id):
        return input(builder,components.InputType.OBSCURED_SHADOW,value,
                     input_block(builder,builder.current_target.sounds[0].name))
    else:
        return input(builder,components.InputType.SHADOW,input_block(builder,value))

def option_input(options: Type[EnumType],shadow: str | Enum,input_block: 'Block') -> FieldOrInput:
    def _option_input(builder: 'ScratchFileBuilder',value: components.Id | str | Enum) -> components.Input:
        if isinstance(value,components.Id):
            return input(builder,components.InputType.OBSCURED_SHADOW,value,input_block(builder,shadow))
        elif isinstance(value,(options,value)):
            return input(builder,components.InputType.SHADOW,input_block(value.value))
        else:
            return input(builder,components.InputType.SHADOW,input_block(value))
    return _option_input

def message_input(builder: 'ScratchFileBuilder',message: components.Id | str) -> components.Input:
    if isinstance(message,components.Id):
        id,name = builder.stage.broadcasts.items()[0]
        return input(builder,components.InputType.OBSCURED_SHADOW,
                     message,components.VariableValue(components.VariableType,name,id))
    else:
        return input(builder,components.InputType.SHADOW,
                     components.VariableValue(components.VariableType.MESSAGE,message,
                        builder._get_message(message)))
    

class Block:
    opcode: str
    mutation: Optional[components.Mutation] = None
    args: list[tuple[str,FieldOrInput]]
    attached_options: EnumType

    def __init__(self,opcode: str,mutation: Optional[components.Mutation]=None,**kwargs: FieldOrInput):
        self.opcode = opcode
        self.mutation = mutation
        self.args = kwargs.items()
        self.attached_options = [None] * len(self.args)

    def attach_options(self,*args: EnumType,**kwargs: EnumType) -> Self:
        self.attached_options[:len(args)] = args
        for name,options in kwargs.items():
            self.attached_options[list(map(lambda arg: arg[0],self.args)).index(name)] = options
        return self
    
    def apply_args(self,builder: 'ScratchFileBuilder',*args: InputItem) -> components.Block:
        inputs = {}
        fields = {}
        for arg,(name,func) in zip(args,self.args):
            arg = func(builder,arg)
            if isinstance(arg,components.Field):
                fields[name] = arg
            elif isinstance(arg,components.Input) and arg[1] is not None:
                inputs[name] = arg
        if self.mutation is None:
            block = components.Block(self.opcode,inputs=inputs,fields=fields)
        else:
            block = components.MutationBlock(self.opcode,inputs=inputs,fields=fields,mutation=self.mutation)
        return block

    def __call__(self,builder: 'ScratchFileBuilder',*args: InputItem,id: Optional[components.Id] = None) -> components.Id:        
        return builder._register_block(self.apply_args(builder,*args),id)


class Monitor:
    opcode: str
    attached_options: Optional[EnumType]

    def __init__(self,opcode: str):
        self.opcode = opcode
        self.attached_options = None

    def attach_options(self,options: EnumType) -> Self:
        self.attached_options = options
        return self

    def __call__(self,builder: 'ScratchFileBuilder',x: int = 0,y: int = 0,visible: bool = True) -> components.Monitor:
        return components.ListMonitor(
            self.opcode.split('_',1)[-1],
            self.opcode,
            x=x,
            y=y,
            visible=visible
        )
    

class SpriteSpecificMonitor(Monitor):
    def __call__(self,builder: 'ScratchFileBuilder',*args,**kwargs) -> components.Monitor:
        monitor = super().__call__(builder,*args,**kwargs)
        if not builder.current_target.isStage:
            monitor.spriteName = builder.current_target.name
        return monitor



class ArchiCatError(Exception):
    def __init__(self,line: int,column: int,message: str):
        super().__init__(f'{message} ({line}:{column})')

class ArchiCatSyntaxError(ArchiCatError): pass

class ArchiCatNameError(ArchiCatError): pass

class ArchiCatInvalidBlockError(ArchiCatError): pass


from . import transpile_file

from argparse import ArgumentParser
from pathlib import Path


def main():
    parser = ArgumentParser(description='ArchiCat CLI')
    parser.add_argument('input',help='input file')
    parser.add_argument('-o','--output',help='output file')
    args = parser.parse_args()
    transpile_file(
        args.input,
        args.output or (Path(args.input).parent / (Path(args.input).stem + '.sb3'))
    )

if __name__ == '__main__':
    main()


from ..block import *
from ..extension import Extension

from enum import Enum
from typing import Optional

class StopOptions(Enum):
    ALL = 'all'
    THIS_SCRIPT = 'this script'
    OTHER_SCRIPTS = 'other scripts in sprite'

class CreateCloneOfOptions(Enum):
    MYSELF = '_myself_'

class StopBlock(Block):
    def __call__(self,builder: 'ScratchFileBuilder',*args: InputItem | StopOptions,id: Optional[components.Id] = None) -> components.Id:
        id = super().__call__(builder,*args,id=id)
        block = builder._block_by_id(id)
        if StopOptions(block.fields['STOP_OPTION'].value) == StopOptions.OTHER_SCRIPTS:
            block.mutation = components.ControlStop()
        else:
            block.mutation = components.ControlStop(hasnext=components.HasNext.FALSE)
        return id

ARCHICAT_EXTENSION = Extension('control')

ARCHICAT_EXTENSION.create_statements(
    Block('control_wait',DURATION=unsigned_float_input),
    Block('control_repeat',TIMES=unsigned_int_input,SUBSTACK=chain_input),
    Block('control_forever',SUBSTACK=chain_input,
          mutation=components.ControlStop(hasnext=components.HasNext.FALSE)),
    Block('control_if',CONDITION=bool_input,SUBSTACK=chain_input),
    Block('control_if_else',CONDITION=bool_input,SUBSTACK=chain_input,SUBSTACK2=chain_input),
    Block('control_wait_until',CONDITION=bool_input),
    Block('control_repeat_until',CONDITION=bool_input,SUBSTACK=chain_input),
    StopBlock('control_stop',STOP_OPTION=option_field(StopOptions)).attach_options(StopOptions),
    Block('control_create_clone_of',CLONE_OPTION=option_input(CreateCloneOfOptions,CreateCloneOfOptions.MYSELF,
            Block('control_create_clone_of_options',CLONE_OPTIONS=option_field(CreateCloneOfOptions)))
    ).attach_options(CreateCloneOfOptions),
    access=BlockAccessModifier.ALL
)

ARCHICAT_EXTENSION.create_statements(
    Block('control_delete_this_clone',
          mutation=components.ControlStop(hasnext=components.HasNext.FALSE))
)

ARCHICAT_EXTENSION.create_hats(
    Block('control_start_as_clone'),
    access=BlockAccessModifier.SPRITE
)


from ..block import *
from ..extension import Extension

from enum import Enum

class GotoOptions(Enum):
    RANDOM = '_random_'
    MOUSE = '_mouse_'

class SetRotationStyleOptions(Enum):
    ALL_AROUND = 'all around'
    LEFT_RIGHT = 'left-right'
    DONT_ROTATE = 'don\'t rotate'

ARCHICAT_EXTENSION = Extension('motion')

ARCHICAT_EXTENSION.create_statements(
    Block('motion_movesteps',STEPS=float_input),
    Block('motion_turnright',DEGREES=float_input),
    Block('motion_turnleft',DEGREES=float_input),
    Block('motion_goto',TO=option_input(GotoOptions,GotoOptions.MOUSE,
                                        Block('motion_goto_menu',TO=option_field(GotoOptions)))).attach_options(GotoOptions),
    Block('motion_gotoxy',X=float_input,Y=float_input),
    Block('motion_glideto',TO=option_input(GotoOptions,GotoOptions.RANDOM,
                                           Block('motion_glideto_menu',TO=option_field(GotoOptions)))).attach_options(GotoOptions),
    Block('motion_pointindirection',DIRECTION=angle_input),
    Block('motion_pointtowards',TO=option_input(GotoOptions,GotoOptions.RANDOM,
                                                Block('motion_pointtowardsmenu',TO=option_field(GotoOptions)))).attach_options(GotoOptions),
    Block('motion_changexby',DX=float_input),
    Block('motion_setx',X=float_input),
    Block('motion_changeyby',DY=float_input),
    Block('motion_sety',Y=float_input),
    Block('motion_setrotationstyle',STYLE=option_field(SetRotationStyleOptions)).attach_options(SetRotationStyleOptions),
    Block('motion_ifonedgebounce'),
    access=BlockAccessModifier.SPRITE,
)

ARCHICAT_EXTENSION.create_reporters(
    Block('motion_xposition'),
    Block('motion_yposition'),
    Block('motion_direction'),
    access=BlockAccessModifier.SPRITE,
)

ARCHICAT_EXTENSION.create_monitors(
    SpriteSpecificMonitor('motion_xposition'),
    SpriteSpecificMonitor('motion_yposition'),
    SpriteSpecificMonitor('motion_direction'),
    access=BlockAccessModifier.SPRITE
)


from ..block import *
from ..extension import Extension

from enum import Enum

class WhenGreaterThanOptions(Enum):
    LOUDNESS = 'LOUDNESS'
    TIMER = 'TIMER'

ARCHICAT_EXTENSION = Extension('event')

ARCHICAT_EXTENSION.create_statements(
    Block('event_broadcast',BROADCAST_INPUT=message_input),
    Block('event_broadcastandwait',BROADCAST_INPUT=message_input),
    access=BlockAccessModifier.ALL,
)

ARCHICAT_EXTENSION.create_hats(
    Block('event_whenflagclicked'),
    Block('event_whenkeypressed',KEY_OPTION=field),
    Block('event_whenbackdropswitchesto',backdrop=costume_field),
    Block('event_whengreaterthan',WHENGREATERTHANMENU=option_field(WhenGreaterThanOptions),VALUE=unsigned_float_input
          ).attach_options(WhenGreaterThanOptions),
    Block('event_whenbroadcastreceived',BROADCAST_OPTION=message_field),
    access=BlockAccessModifier.ALL,
)

ARCHICAT_EXTENSION.create_hats(
    Block('event_whenthisspriteclicked'),
    access=BlockAccessModifier.SPRITE,
)

ARCHICAT_EXTENSION.create_hats(
    Block('event_whenstageclicked'),
    access=BlockAccessModifier.STAGE,
)


from ..block import *
from ..extension import Extension

from enum import Enum

class TouchingObjectOptions(Enum):
    MOUSE = '_mouse_'
    EDGE = '_edge_'

class DistanceToOptions(Enum):
    MOUSE = '_mouse_'

class KeyPressedOptions(Enum):
    SPACE = 'space'
    LEFT_ARROW = 'left arrow'
    RIGHT_ARROW = 'right arrow'
    UP_ARROW = 'up arrow'
    DOWN_ARROW = 'down arrow'
    ANY = 'any'

class SetDragModeOptions(Enum):
    DRAGGABLE = 'draggable'
    NOT_DRAGGABLE = 'not draggable'

class OfPropertyOptions(Enum):
    X = 'x position'
    Y = 'y position'
    DIRECTION = 'direction'
    COSTUME_NUMBER = 'costume #'
    COSTUME_NAME = 'costume name'
    BACKDROP_NUMBER = 'backdrop #'
    BACKDROP_NAME = 'backdrop name'
    SIZE = 'size'
    VOLUME = 'volume'

class OfObjectOptions(Enum):
    STAGE = '_stage_'

class CurrentOptions(Enum):
    YEAR = 'YEAR'
    MONTH = 'MONTH'
    DATE = 'DATE'
    DAY_OF_WEEK = 'DAYOFWEEK'
    HOUR = 'HOUR'
    MINUTE = 'MINUTE'
    SECOND = 'SECOND'


class CurrentMonitor(Monitor):
    def __call__(self,builder: 'ScratchFileBuilder',option: str | CurrentOptions,**kwargs) -> components.ListMonitor:
        monitor = super().__call__(builder,**kwargs)
        option = CurrentOptions(option)
        monitor.params = {'CURRENTMENU': option}
        monitor.id += '_' + option.value.lower()
        return monitor

ARCHICAT_EXTENSION = Extension('sensing')

ARCHICAT_EXTENSION.create_statements(
    Block('sensing_askandwait',QUESTION=string_input),
    Block('sensing_resettimer'),
    access=BlockAccessModifier.ALL
)

ARCHICAT_EXTENSION.create_statements(
    Block('sensing_setdragmode',DRAG_MODE=option_field(SetDragModeOptions)).attach_options(SetDragModeOptions),
    access=BlockAccessModifier.SPRITE
)

ARCHICAT_EXTENSION.create_reporters(
    Block('sensing_answer'),
    Block('sensing_keypressed',KEY_OPTION=option_input(KeyPressedOptions,KeyPressedOptions.SPACE,
                                Block('sensing_keyoptions',KEY_OPTION=option_field(KeyPressedOptions)))).attach_options(KeyPressedOptions),
    Block('sensing_mousedown'),
    Block('sensing_mousex'),
    Block('sensing_mousey'),
    Block('sensing_loudness'),
    Block('sensing_timer'),
    Block('sensing_of',OBJECT=option_input(OfObjectOptions,OfObjectOptions.STAGE,
                        Block('sensing_of_object_menu',OBJECT=option_field(OfObjectOptions))),
                        PROPERTY=option_field(OfPropertyOptions)).attach_options(OfObjectOptions),
    Block('sensing_current',CURRENT_MENU=option_field(CurrentOptions)).attach_options(CurrentOptions),
    Block('sensing_dayssince2000'),
    Block('sensing_username'),
    access=BlockAccessModifier.ALL
)

ARCHICAT_EXTENSION.create_reporters(
    Block('sensing_touchingobject',TOUCHINGOBJECTMENU=option_input(TouchingObjectOptions,TouchingObjectOptions.MOUSE,
                                    Block('sensing_touchingobjectmenu',TOUCHINGOBJECTMENU=option_field(TouchingObjectOptions)))
            ).attach_options(TouchingObjectOptions),
    Block('sensing_touchingcolor',COLOR=color_input),
    Block('sensing_coloristouchingcolor',COLOR=color_input,COLOR2=color_input),
    Block('sensing_distanceto',DISTANCETOMENU=option_input(DistanceToOptions,DistanceToOptions.MOUSE,
                                Block('sensing_distancetomenu',DISTANCETOMENU=option_field(DistanceToOptions)))
            ).attach_options(DistanceToOptions),
    access=BlockAccessModifier.SPRITE
)

ARCHICAT_EXTENSION.create_monitors(
    Monitor('sensing_username'),
    Monitor('sensing_answer'),
    Monitor('sensing_loudness'),
    Monitor('sensing_timer'),
    access=BlockAccessModifier.ALL,
)


from .transformer import ScratchFileBuilder
from .parser import parse

from pathlib import Path


def transpile(source: str,target: Path | str):
    target = str(target)
    builder = ScratchFileBuilder()
    builder.visit(parse(source))
    builder.save(target)

def transpile_file(source: Path | str,target: Path | str):
    with open(str(source)) as file:
        transpile(file.read(),target)

# TODO replace types with strings for all TYPE_CHECKING imports


from ..block import *
from ..extension import Extension


class VariableMonitor(Monitor):
    def __call__(self,builder: 'ScratchFileBuilder',variable: str,mode: str = 'default',
                 sliderMin: int = 0,sliderMax: int = 100,isDiscrete: bool = True,**kwargs) -> components.Monitor:
        monitor = super().__call__(builder,**kwargs)
        monitor.id = builder._get_variable(variable)
        monitor.params = {'VARIABLE': variable}
        if not builder.current_target.isStage:
            monitor.spriteName = builder.current_target.name
        monitor.mode = mode
        monitor.sliderMin = sliderMin
        monitor.sliderMax = sliderMax
        monitor.isDiscrete = isDiscrete
        return monitor
    
class ListMonitor(Monitor):
    def __call__(self,builder: 'ScratchFileBuilder',list: str,width: int = 100,height: int = 200,**kwargs) -> components.ListMonitor:
        monitor = super().__call__(builder,**kwargs)
        monitor.id = builder._get_list(list)
        monitor.params = {'LIST': list}
        if not builder.current_target.isStage:
            monitor.spriteName = builder.current_target.name
        monitor.mode = 'list'
        monitor.value = []
        monitor.width = width
        monitor.height = height
        return monitor

ARCHICAT_EXTENSION = Extension('data')

ARCHICAT_EXTENSION.create_statements(
    Block('data_setvariableto',VARIABLE=variable_field,VALUE=string_input),
    Block('data_changevariableby',VARIABLE=variable_field,VALUE=float_input),
    Block('data_showvariable',VARIABLE=variable_field),
    Block('data_hidevariable',VARIABLE=variable_field),
    Block('data_addtolist',LIST=list_field,ITEM=string_input),
    Block('data_deleteoflist',LIST=list_field,INDEX=unsigned_int_input),
    Block('data_deletealloflist',LIST=list_field),
    Block('data_insertatlist',LIST=list_field,INDEX=unsigned_int_input,ITEM=string_input),
    Block('data_replaceitemoflist',LIST=list_field,INDEX=unsigned_int_input,ITEM=string_input),
    Block('data_showlist',LIST=list_field),
    Block('data_hidelist',LIST=list_field),
    access=BlockAccessModifier.ALL
)

ARCHICAT_EXTENSION.create_reporters(
    Block('data_variable',VARIABLE=variable_field),
    Block('data_listcontents',LIST=list_field),
    Block('data_itemoflist',LIST=list_field,INDEX=unsigned_int_input),
    Block('data_itemnumoflist',LIST=list_field,ITEM=string_input),
    Block('data_lengthoflist',LIST=list_field),
    Block('data_listcontainsitem',LIST=list_field,ITEM=string_input),
    access=BlockAccessModifier.ALL
)

ARCHICAT_EXTENSION.create_monitors(
    VariableMonitor('data_variable'),
    ListMonitor('data_listcontents')
)


from .block import *

from pathlib import Path
from pkgutil import iter_modules
from importlib import import_module
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .transformer import ScratchFileBuilder

type ExtensionBlockDict = dict[str,tuple[Block,BlockAccessModifier]]


class Extension:
    name: str
    statement_blocks: ExtensionBlockDict
    reporter_blocks: ExtensionBlockDict
    hat_blocks: ExtensionBlockDict
    monitors: dict[str,tuple[Monitor,BlockAccessModifier]]

    def __init__(self,name: str = ''):
        self.name = name.lower()
        self.statement_blocks = {}
        self.reporter_blocks = {}
        self.hat_blocks = {}
        self.monitors = {}

    def _format_name(self,name: str) -> str:
        return name.lower().removeprefix(self.name + '_').replace('_','')

    def create_statement(self,name: str,block: Block,access: BlockAccessModifier = BlockAccessModifier.ALL):
        self.statement_blocks[self._format_name(name)] = (block,access)

    def create_reporter(self,name: str,block: Block,access: BlockAccessModifier = BlockAccessModifier.ALL):
        self.reporter_blocks[self._format_name(name)] = (block,access)

    def create_hat(self,name: str,block: Block,access: BlockAccessModifier = BlockAccessModifier.ALL):
        self.hat_blocks[self._format_name(name)] = (block,access)

    def create_monitor(self,name: str,monitor: Monitor,access: BlockAccessModifier = BlockAccessModifier.ALL):
        self.monitors[self._format_name(name)] = (monitor,access)

    def create_statements(self,*blocks: Monitor,access: BlockAccessModifier = BlockAccessModifier.ALL):
        for block in blocks:
            self.create_statement(block.opcode,block,access)

    def create_reporters(self,*blocks: Block,access: BlockAccessModifier = BlockAccessModifier.ALL):
        for block in blocks:
            self.create_reporter(block.opcode,block,access)

    def create_hats(self,*blocks: Block,access: BlockAccessModifier = BlockAccessModifier.ALL):
        for block in blocks:
            self.create_hat(block.opcode,block,access)

    def create_monitors(self,*monitors: Monitor,access: BlockAccessModifier = BlockAccessModifier.ALL):
        for monitor in monitors:
            self.create_monitor(monitor.opcode,monitor,access)

    def has_statement(self,name: str,context: BlockAccessModifier) -> bool:
        return self._format_name(name) in self.statement_blocks and \
            ((access := self.statement_blocks[self._format_name(name)][1]) == context or access == BlockAccessModifier.ALL)
    
    def has_reporter(self,name: str,context: BlockAccessModifier) -> bool:
        return self._format_name(name) in self.reporter_blocks and \
            ((access := self.reporter_blocks[self._format_name(name)][1]) == context or access == BlockAccessModifier.ALL)
    
    def has_hat(self,name: str,context: BlockAccessModifier) -> bool:
        return self._format_name(name) in self.hat_blocks and \
            ((access := self.hat_blocks[self._format_name(name)][1]) == context or access == BlockAccessModifier.ALL)
    
    def has_monitor(self,name: str,context: BlockAccessModifier) -> bool:
        return self._format_name(name) in self.monitors and \
            ((access := self.monitors[self._format_name(name)][1]) == context or access == BlockAccessModifier.ALL)

    def get_statement(self,name: str) -> Block:
        return self.statement_blocks[self._format_name(name)][0]
    
    def get_reporter(self,name: str) -> Block:
        return self.reporter_blocks[self._format_name(name)][0]
    
    def get_hat(self,name: str) -> Block:
        return self.hat_blocks[self._format_name(name)][0]
    
    def get_monitor(self,name: str) -> Monitor:
        return self.monitors[self._format_name(name)][0]
    

class ExtensionManager:
    src: str
    extensions: list[Extension]
    builder: 'ScratchFileBuilder'

    def __init__(self,builder: 'ScratchFileBuilder',src: str = 'archicat.blocks'):
        self.src = src
        self.builder = builder
        self.extensions = []
        
        for module in iter_modules(import_module(self.src).__path__):
            if not module.ispkg:
                self.extensions.append(import_module(f'{self.src}.{module.name}').ARCHICAT_EXTENSION)

    def load_extension(self,name: str):
        self.extensions.append(extension := import_module(f'{self.src}.extensions.{name}').ARCHICAT_EXTENSION)
        self.builder.project.extensions.append(extension.name)

    def add_extension(self,extension: Extension):
        self.extensions.append(extension)

    def update(self,name: str,extension: Extension):
        new = ExtensionManager(self.builder,self.src)
        for new_extension in self.extensions:
            if new_extension.name == name:
                new.add_extension(extension)
            else:
                new.add_extension(new_extension)
        return new

    def get_extension(self,name: str) -> Extension:
        return next(extension for extension in self.extensions if extension.name == name)

    def get_statement(self,name: str) -> Block:
        for extension in self.extensions:
            if extension.has_statement(name,BlockAccessModifier.STAGE 
                                       if self.builder.current_target.isStage else 
                                       BlockAccessModifier.SPRITE):
                return extension.get_statement(name)
            
    def get_reporter(self,name: str) -> Block:
        for extension in self.extensions:
            if extension.has_reporter(name,BlockAccessModifier.STAGE 
                                       if self.builder.current_target.isStage else 
                                       BlockAccessModifier.SPRITE):
                return extension.get_reporter(name)
            
    def get_hat(self,name: str) -> Block:
        for extension in self.extensions:
            if extension.has_hat(name,BlockAccessModifier.STAGE 
                                       if self.builder.current_target.isStage else 
                                       BlockAccessModifier.SPRITE):
                return extension.get_hat(name)
            
    def get_monitor(self,name: str) -> Monitor:
        for extension in self.extensions:
            if extension.has_monitor(name,BlockAccessModifier.STAGE
                                        if self.builder.current_target.isStage else
                                        BlockAccessModifier.SPRITE):
                return extension.get_monitor(name)
    



from .error import ArchiCatSyntaxError

from lark import Lark,Tree
from lark.exceptions import UnexpectedEOF,UnexpectedCharacters,UnexpectedToken
from pathlib import Path
from importlib.resources import files


parser = Lark(files('archicat').joinpath('grammar.lark').read_text())

def parse(text: str) -> Tree:
    try:
        return parser.parse(text)
    except UnexpectedToken as exception:
        raise ArchiCatSyntaxError(exception.line,exception.column,f'Unexpected token {exception.token}')
    except UnexpectedCharacters as exception:
        raise ArchiCatSyntaxError(exception.line,exception.column,f'Unexpected character {exception.char}')
    except UnexpectedEOF as exception:
        raise ArchiCatSyntaxError(exception.line,exception.column,'Unexpected EOF')

def parse_file(path: Path | str) -> Tree:
    with open(str(path)) as file:
        return parse(file.read())


from archicat import components
from .block import Block,bool_input,string_input
from .extension import ExtensionManager,Extension

from lark import Tree
from lark.visitors import Interpreter,v_args
from zipfile import ZipFile
from pathlib import Path
from typing import Optional,Type,Callable
from uuid import uuid4
from enum import Enum
from json import dumps as dump_json,loads as load_json
from hashlib import md5
from wave import open as open_wav
from mutagen.mp3 import MP3
from enum import Enum,auto


def random_id(name: str = '') -> components.Id:
    return components.Id(name + '_' * bool(len(name)) + uuid4().hex)


class IdentifierType(Enum):
    VARIABLE = 'variable'
    LIST = 'list'
    MESSAGE = 'message'

class ArgumentType(Enum):
    STRING = auto()
    BOOL = auto()


@v_args(inline=True)
class ScratchFileBuilder(Interpreter):
    assets: list[tuple[str,Path]]
    project: components.Project
    stage: components.Stage
    current_target: components.Target
    block_stack: list[ExtensionManager]
    expands_file: Optional[Path] = None
    parent_block_stack: list[components.Id]
    identifier_type_stack: list[IdentifierType]
    available_options_stack: list[Type[Enum]]
    secondary_callbacks: list[tuple[Callable,components.Target,Extension]]

    def __init__(self):
        self.assets = []
        self.stage = components.Stage('Stage',True)
        self.current_target = self.stage
        self.block_stack = [ExtensionManager(self)]
        self.block_stack[-1].add_extension(Extension('custom'))
        self.parent_block_stack = [None]
        self.identifier_type_stack = []
        self.available_options_stack = []
        self.secondary_callbacks = []
        self.expands_file = None
        self.project = components.Project(targets=[self.stage])

    def save(self,path: Path | str):
        for callback,self.current_target,self.block_stack in self.secondary_callbacks:
            callback()
        with ZipFile(str(path),'w') as zipfile:
            zipfile.writestr('project.json',data := dump_json(components.to_json(self.project),indent=4))
            # print(data)
            for name,path in self.assets:
                zipfile.writestr(name,path.read_bytes())
            if self.expands_file is not None:
                with ZipFile(self.expands_file,'r') as srcfile:
                    for asset in srcfile.namelist():
                        if asset != 'project.json':
                            zipfile.writestr(asset,srcfile.read(asset))

    def _secondary_decorator(self,callback: Callable):
        self.secondary_callbacks.append((callback,self.current_target,self.block_stack.copy()))

    def _get_variable(self,name: str) -> components.Id:
        return next(id for id,variable in (*self.current_target.variables.items(),
                    *self.stage.variables.items()) if variable.name == name)
    
    def _get_list(self,name: str) -> components.Id:
        return next(id for id,list in (*self.current_target.lists.items(), 
                    *self.stage.lists.items()) if list.name == name)
    
    def _get_message(self,name: str) -> components.Id:
        return next(id for id,msgname in self.stage.broadcasts.items() 
                    if name == msgname)
    
    def _block_by_id(self,id: components.Id) -> components.Block:
        return self.current_target.blocks[id]
    
    def _register_block(self,block: components.Block | components.VariableValue,id: Optional[components.Id] = None) -> components.Id:
        self.current_target.blocks[id or (id := random_id())] = block
        return id
    
    # def _statement_blocks(self) -> dict[str,Block]:
    #     return (stage_statement_blocks if self.current_target.isStage else sprite_statement_blocks) | self.procedure_scope_stack[-1]
    
    # def _reporter_blocks(self) -> dict[str,Block]:
    #     return stage_reporter_blocks if self.current_target.isStage else sprite_reporter_blocks
    
    # def _hat_blocks(self) -> dict[str,Block]:
    #     return stage_hat_blocks if self.current_target.isStage else sprite_hat_blocks
    
    def _transform_block(self,block: Block,comment: Tree | None,*args: Tree) -> components.Id:
        self.parent_block_stack.append(block_id := random_id())
        processed_args = []
        for arg,arg_option in zip(args,block.attached_options):
            self.available_options_stack.append(arg_option)
            processed_args.append(self.visit(arg))
            self.available_options_stack.pop()
        block(self,*processed_args,id=block_id)
        if comment is not None:
            self.current_target.comments[comment_id := self.visit(comment)].blockId = block_id
            self.current_target.blocks[block_id] = components.CommentedBlock(**vars(self._block_by_id(block_id)),comment=comment_id)
        return self.parent_block_stack.pop()
    
    def _create_procedure(self,procname: str,chain: Tree,position: Tree,*arguments: Tree,warp: bool):
        x,y = self.visit(position)
        arguments = list(map(self.visit,arguments))
        argument_names = []
        argument_types = []
        argument_ids = []
        argument_defaults = []
        argument_blocks = []
        prototype_id = random_id()
        for name,type in arguments:
            argument_names.append(name)
            argument_types.append(type)
            argument_ids.append(random_id())
            argument_defaults.append('false' if type == ArgumentType.BOOL else '')
            if type == ArgumentType.STRING:
                argument_blocks.append(self._register_block(
                    components.Block('argument_reporter_string_number',fields = {'VALUE': components.Field(name)},shadow=True,parent=prototype_id)))
            else:
                argument_blocks.append(self._register_block(
                    components.Block('argument_reporter_boolean',fields = {'VALUE': components.Field(name)},shadow=True,parent=prototype_id)))
        definition_id = self._register_block(components.HatBlock('procedures_definition',inputs = {
                'custom_block': components.Input(components.InputType.SHADOW,prototype_id)
            },topLevel=True,x=x,y=y))
        self._register_block(components.MutationBlock('procedures_prototype',inputs = {
                id: components.Input(components.InputType.SHADOW,argument_block) for id,argument_block in zip(argument_ids,argument_blocks)
            },shadow=True,parent=definition_id,mutation=components.ProcedurePrototype(
                proccode=(proccode := procname + ' ' + ' '.join('%b' if type == ArgumentType.BOOL else '%s' for _,type in arguments)),
                argumentids=dump_json(argument_ids),
                argumentnames=dump_json(argument_names),
                argumentdefaults=dump_json(argument_defaults),
                warp=components.Warp.TRUE if warp else components.Warp.FALSE
            )),id=prototype_id)
        
        self.block_stack[-1].get_extension('custom').create_statement(procname,
            Block('procedures_call',components.Procedure(
                proccode=proccode,
                argumentids=dump_json(argument_ids),
                warp=components.Warp.TRUE if warp else components.Warp.FALSE
            ),**{argument_id: (bool_input if argument_type == ArgumentType.BOOL else string_input) 
                            for argument_id,argument_type in zip(argument_ids,argument_types)}))
        
        @self._secondary_decorator
        def secondary():
            self._block_by_id(definition_id).next = chain_id = self.visit(chain)
            if chain_id is not None:
                self._block_by_id(chain_id).parent = definition_id

    def start(self,*declarations):
        self.current_target = self.stage
        for declaration in declarations:
            self.visit(declaration)
        return self.project

    def sprite(self,name,*declarations):
        self.block_stack.append(self.block_stack[-1].update('custom',Extension('custom')))
        if targets := list(target for target in self.project.targets if target.name == name):
            self.current_target = targets[0]
        else:
            self.current_target = components.Sprite(name)
            self.project.targets.append(self.current_target)
        for declaration in declarations:
            self.visit(declaration)
        self.current_target = self.stage
        self.block_stack.pop()

    def expand(self,file):
        self.expands_file = Path(file[1:-1])
        with ZipFile(str(self.expands_file)) as zipfile:
            self.project = components.from_json(load_json(zipfile.read('project.json')))
            self.stage = next(target for target in self.project.targets if target.isStage)
            self.current_target = self.stage

    def config_option(self,name,value):
        setattr(self.current_target,name,self.visit(value))

    def config(self,*options):
        for option in options:
            self.visit(option)

    def monitor_option(self,name,value):
        return name,self.visit(value)
    
    def monitor(self,name,arg,*options):
        monitor = self.block_stack[-1].get_monitor(name)
        self.available_options_stack.append(monitor.attached_options)
        if arg is None:
            self.project.monitors.append(monitor(self,**dict(map(self.visit,options))))
        else:
            self.project.monitors.append(monitor(self,self.visit(arg),**dict(map(self.visit,options))))
        self.available_options_stack.pop(0)

    def comment(self,content: str,*options: Tree) -> components.Id:
        self.current_target.comments[id := random_id()] = components.Comment(text=content[1:-1],**dict(map(self.visit,options)))
        return id

    def variable(self,name,value=None):
        self.current_target.variables[random_id()] = \
            components.Variable(name.value,self.visit(value) if value is not None else '')

    def list(self,name,*values):
        self.current_target.lists[random_id()] = \
            components.List(name.value,list(map(self.visit,values)))
        
    def message(self,name):
        self.stage.broadcasts[random_id()] = name.value

    def costume(self,name,path):
        path = Path(path[1:-1])
        hash = md5(path.read_bytes()).hexdigest()
        suffix = path.suffix.lstrip('.').lower()
        self.current_target.costumes.append(components.Costume(
            assetId=hash,
            name=name,
            md5ext=hash + '.' + suffix,
            dataFormat=suffix,
            bitmapResolution=1,
            rotationCenterX=0,
            rotationCenterY=0
        ))
        self.assets.append((hash + '.' + suffix,path))

    def sound(self,name,path):
        path = Path(path[1:-1])
        hash = md5(path.read_bytes()).hexdigest()
        suffix = path.suffix.lstrip('.').lower()
        if path.suffix.lstrip('.').lower() == 'wav':
            with open_wav(str(path)) as wav:
                rate = wav.getframerate()
                sample_count = wav.getnframes()
        elif path.suffix.lstrip('.').lower() == 'mp3':
            audio = MP3(path)
            rate = audio.info.sample_rate
            sample_count = int(round(rate * audio.info.length))
            
        self.current_target.sounds.append(components.Sound(
            assetId=hash,
            name=name,
            md5ext=hash + '.' + suffix,
            dataFormat=suffix,
            rate=rate,
            sampleCount=sample_count
        ))
        self.assets.append((hash + '.' + suffix,path))

    def option(self,name):
        if self.available_options_stack[-1] is not None:
            return self.available_options_stack[-1]._member_map_[name]
        raise NameError(f'Option {name} not found')
        
    def identifier(self,name): return name

    def event(self,hat,position,chain):
        @self._secondary_decorator
        def secondary():
            nonlocal hat,chain
            x,y = self.visit(position)
            hat = self.visit(hat)
            chain = self.visit(chain)
            self._block_by_id(hat).topLevel = True
            self._block_by_id(hat).next = chain
            self._block_by_id(hat).x = x
            self._block_by_id(hat).y = y
            if chain is not None:
                self._block_by_id(chain).parent = hat

    def procedure(self,name,*args):
        *args,position,chain = args
        self._create_procedure(name,chain,position,*args,warp=False)

    def warp_procedure(self,name,*args):
        *args,position,chain = args
        self._create_procedure(name,chain,position,*args,warp=True)

    def string_argument(self,name):
        return name,ArgumentType.STRING
    
    def bool_argument(self,name):
        return name,ArgumentType.BOOL
    
    def position(self,x=0,y=0):
        return int(x),int(y)

    def chain(self,*blocks):
        blocks = list(map(self.visit,blocks))
        for n,block in enumerate(blocks):
            if n > 0: self._block_by_id(blocks[n - 1]).next = block
            if n + 1 < len(blocks): self._block_by_id(blocks[n + 1]).parent = block
        if len(blocks):
            return blocks[0]

    def statement_block(self,name,*args):
        *args,comment = args
        return self._transform_block(self.block_stack[-1].get_statement(name),comment,*args)
    
    def reporter_block(self,name,*args):
        *args,comment = args
        return self._transform_block(self.block_stack[-1].get_reporter(name),comment,*args)
    
    def hat_block(self,name,*args):
        *args,comment = args
        return self._transform_block(self.block_stack[-1].get_hat(name),comment,*args)

    def int(self,value):
        return int(value)
    
    def float(self,value):
        return float(value)
    
    def string(self,value):
        return str(value[1:-1])
    
    def color(self,value):
        return str(value)
    
    def true(self):
        return True
    
    def false(self):
        return False


def transform(tree: Tree) -> components.Project:
    builder = ScratchFileBuilder()
    builder.visit(tree)
    return builder


from pydantic.dataclasses import dataclass
from dataclasses import field
from typing import Any,Optional


class Writer:
    def generate_code(self) -> str: pass

class Chain[T = Writer](list[T]):
    def generate_code(self) -> str:
        return '{' + '\n'.join(self) + '}'
    
class Declaration(Writer): pass
    
class Value(Writer):
    value: str

    def __init__(self,value: str | int | float | bool):
        if isinstance(value,bool):
            self.value = 'true' if self.value else 'false'
        else:
            self.value = repr(value)

    def generate_code(self):
        return self.value
    
class Path(Writer):
    path: str

    def generate_code(self):
        return '"' + self.path + '"'

@dataclass
class ListValue(Writer):
    values: list[Value] = field(default_factory=list)

    def generate_code(self):
        return f'[{','.join(map(to_code,self.values))}]'
    
@dataclass
class Identifier(Writer):
    name: str

    def generate_code(self):
        return self.name    
    
@dataclass
class Option(Identifier):
    def generate_code(self):
        return '$' + super().generate_code()
    
class Options(Writer):
    options: dict[str,Value]

    def __init__(self,**options):
        self.options = options

    def generate_code(self):
        return '{' + '\n'.join([f'{name} = {value}' for name,value in self.options.items()]) + '}'
    
@dataclass
class BlockCall(Writer):
    name: str
    args: tuple[Value | 'BlockCall' | Chain | Identifier,...] = ()

    def generate_code(self):
        return f'{self.name}({','.join(map(to_code,self.args))})'
    
@dataclass
class Comment(Declaration):
    content: str
    options: Options
    
    def generate_code(self):
        return f': "{self.content}" {to_code(self.options)}'
    
@dataclass
class VariableDeclaration(Declaration):
    name: str
    value: Optional[Value] = None

    def generate_code(self):
        if self.value is None:
            return f'var {self.name}'
        else:
            return f'var {self.name} = {to_code(self.value)}'
        
@dataclass
class ListDeclaration(Declaration):
    name: str
    value: ListValue = field(default_factory=ListValue)

    def generate_code(self):
        return f'list {self.name} = {to_code(self.value)}'
    
@dataclass
class MessageDeclaration(Declaration):
    name: str

    def generate_code(self):
        return f'message {self.name}'
    
@dataclass
class CostumeDeclaration(Declaration):
    name: str
    path: Path

    def generate_code(self):
        return f'costume {self.name} = {to_code(self.path)}'
    
@dataclass
class SoundDeclaration(Declaration):
    name: str
    path: Path

    def generate_code(self):
        return f'sound {self.name} = {to_code(self.path)}'
    
@dataclass
class ProcedureDeclaration(Declaration):
    name: str
    args: tuple[str,...] = ()
    position: Optional[tuple[int,int]] = None
    warp: bool = False

    def generate_code(self):
        if self.warp:
            return f'warp {self.name}({','.join(self.args)})' \
                + ('' if self.position is None else f'[{self.position[0]},{self.position[1]}]')
        else:
            return f'proc {self.name}({','.join(self.args)})' \
                + ('' if self.position is None else f'[{self.position[0]},{self.position[1]}]')
        
@dataclass
class Event(Declaration):
    block: BlockCall
    chain: Chain
    position: Optional[tuple[int,int]] = None

    def generate_code(self):
        return f'on {to_code(self.block)} {to_code(self.chain)}' \
            + ('' if self.position is None else f'[{self.position[0]},{self.position[1]}]')
    
@dataclass
class Configuration(Declaration):
    def generate_code(self):
        return f'config {super().generate_code()}'
    
@dataclass
class Target(Writer):
    name: str
    declarations: Chain[Declaration] = field(default_factory=Chain[Declaration])

    def generate_code(self):
        return f'sprite {self.name} {to_code(self.declarations)}'   



def to_code(writer: Writer) -> str:
    return writer.generate_code()


from ..block import *
from ..extension import Extension

from enum import Enum

class EffectOptions(Enum):
    PITCH = 'PITCH'
    PAN = 'PAN'

ARCHICAT_EXTENSION = Extension('sound')

ARCHICAT_EXTENSION.create_statements(
    Block('sound_playuntildone',SOUND_MENU=sound_input),
    Block('sound_play',SOUND_MENU=sound_input),
    Block('sound_changeeffectby',EFFECT=option_field(EffectOptions),CHANGE=float_input).attach_options(EffectOptions),
    Block('sound_seteffectto',EFFECT=option_field(EffectOptions),VALUE=unsigned_float_input).attach_options(EffectOptions),
    Block('sound_cleareffects'),
    Block('sound_changevolumeby',VOLUME=float_input),
    Block('sound_setvolumeto',VOLUME=float_input),
    access=BlockAccessModifier.ALL,
)

ARCHICAT_EXTENSION.create_reporters(
    Block('sound_volume'),
    access=BlockAccessModifier.ALL,
)

ARCHICAT_EXTENSION.create_monitors(
    SpriteSpecificMonitor('sound_volume'),
    access=BlockAccessModifier.ALL
)


from ..block import *
from ..extension import Extension

from enum import Enum

def _operation(opcode):
    return Block(opcode,NUM1=float_input,NUM2=float_input)

def _comparison(opcode):
    return Block(opcode,OPERAND1=float_input,OPERAND2=float_input)

def _bool_operation(opcode):
    return Block(opcode,OPERAND1=bool_input,OPERAND2=bool_input)

class MathOpOptions(Enum):
    ABS = 'abs'
    FLOOR = 'floor'
    CEILING = 'ceiling'
    SQRT = 'sqrt'
    SIN = 'sin'
    COS = 'cos'
    TAN = 'tan'
    ASIN = 'asin'
    ACOS = 'acos'
    ATAN = 'atan'
    POWER_E = 'e ^'
    POWER_10 = '10 ^'

ARCHICAT_EXTENSION = Extension('operator')

ARCHICAT_EXTENSION.create_reporters(
    _operation('operator_add'),
    _operation('operator_subtract'),
    _operation('operator_multiply'),
    _operation('operator_divide'),
    Block('operator_random',FROM=float_input,TO=float_input),
    _comparison('operator_equals'),
    _comparison('operator_gt'),
    _comparison('operator_lt'),
    _bool_operation('operator_and'),
    _bool_operation('operator_or'),
    Block('operator_not',OPERAND=bool_input),
    Block('operator_join',STRING1=string_input,STRING2=string_input),
    Block('operator_letter_of',STRING=string_input),
    Block('operator_length',STRING=string_input),
    _operation('operator_mod'),
    Block('operator_round',NUM=float_input),
    Block('operator_mathop',OPERATION=option_field(MathOpOptions),NUM=float_input).attach_options(MathOpOptions),
    access=BlockAccessModifier.ALL
)


from ..block import *
from ..extension import Extension

ARCHICAT_EXTENSION = Extension('arguments')

ARCHICAT_EXTENSION.create_reporters(
    Block('argument_reporter_boolean',VALUE=field),
    Block('argument_reporter_string_number',VALUE=field),
    access=BlockAccessModifier.ALL
)


from typing import Any,Union,Literal,NamedTuple,Optional
from pydantic.dataclasses import dataclass
from pydantic_core import core_schema
from dataclasses import field,asdict,is_dataclass
from enum import Enum

class Id(str):
    @classmethod
    def __get_pydantic_core_schema__(cls,*args,**kwargs):
        return core_schema.no_info_after_validator_function(cls,core_schema.str_schema())

    def __repr__(self):
        return f'Id(\'{self}\')'
    
# type Id = str


class InputType(Enum):
    SHADOW = 1
    NO_SHADOW = 2
    OBSCURED_SHADOW = 3

class ValueType(Enum):
    FLOAT = 4
    UNSIGNED_FLOAT = 5
    UNSIGNED_INT = 6
    INT = 7
    ANGLE = 8
    COLOR = 9
    STRING = 10

class VariableType(Enum):
    MESSAGE = 11
    VARIABLE = 12
    LIST = 13

class VideoState(Enum):
    ON = 'on'
    OFF = 'off'
    FLIPPED = 'on-flipped'

class RotationStyle(Enum):
    ALL_AROUND = 'all around'
    LEFT_RIGHT = 'left-right'
    DONT_ROTATE = 'don\'t rotate'

class Warp(Enum):
    TRUE = 'true'
    FALSE = 'false'

class HasNext(Enum):
    TRUE = 'true'
    FALSE = 'false'



class Value(NamedTuple):
    type: Union[ValueType,VariableType]
    value: Any
    
class VariableValue(NamedTuple):
    type: VariableType
    value: Any
    id: Id
    x: Optional[int] = None
    y: Optional[int] = None

class Input(NamedTuple):
    type: InputType
    input: Union[Id,Value,VariableValue]
    shadow: Optional[Union[Id,Value,VariableValue]] = None
    
class Field(NamedTuple):
    value: Any
    id: Optional[Id] = None

class Variable(NamedTuple):
    name: str
    value: Any = ''
    cloud: bool = False
    
class List(NamedTuple):
    name: str
    value: list[Any] = field(default_factory = list)

@dataclass
class ListMonitor:
    id: Id
    opcode: str
    params: dict[str,Any] = field(default_factory=dict)
    mode: str = 'default'
    spriteName: Optional[str] = None
    value: Any = None
    width: int = 0
    height: int = 0
    x: int = 0
    y: int = 0
    visible: bool = True
    
@dataclass
class Monitor(ListMonitor):
    sliderMin: int = 0
    sliderMax: int = 100
    isDiscrete: bool = True
    
@dataclass
class Comment:
    text: str
    blockId: Optional[Id] = None
    minimized: bool = False
    x: int = 0
    y: int = 0
    width: int = 100
    height: int = 100
    
@dataclass
class Meta:
    semver: str = '3.0.0'
    vm: str = '5.0.0'
    agent: str = ''

@dataclass
class Block:
    opcode: str
    inputs: dict[str,Input] = field(default_factory = dict)
    fields: dict[str,Field] = field(default_factory = dict)
    next: Optional[Id] = None
    parent: Optional[Id] = None
    shadow: bool = False 
    topLevel: bool = False

@dataclass
class CommentedBlock(Block):
    comment: Id = Id()

@dataclass
class HatBlock(Block):
    x: int = 0
    y: int = 0

@dataclass
class MutationBlock(Block):
    mutation: Optional['Mutation'] = None
    
@dataclass
class Mutation:
    tagName: str = 'mutation'
    children: list[Any] = field(default_factory = list)
    
@dataclass
class Procedure(Mutation):
    proccode: str = ''
    argumentids: str = ''
    warp: Warp | str = Warp.FALSE
    
@dataclass
class ProcedurePrototype(Procedure):
    argumentnames: str = ''
    argumentdefaults: str = ''
    
@dataclass
class ControlStop(Mutation):
    hasnext: HasNext | str = HasNext.TRUE
    
@dataclass
class Asset:
    assetId: str = ''
    name: str = ''
    md5ext: str = ''
    dataFormat: str = ''
    
@dataclass
class Costume(Asset):
    bitmapResolution: int = 1
    rotationCenterX: float = 0
    rotationCenterY: float = 0
    
@dataclass
class Sound(Asset):
    rate: float = 1
    sampleCount: int = 1
    
@dataclass
class Target:
    name: str
    isStage: bool = False
    variables: dict[Id,Variable] = field(default_factory = dict)
    lists: dict[Id,List] = field(default_factory = dict)
    broadcasts: dict[Id,str] = field(default_factory = dict)
    blocks: dict[Id,Block | VariableValue] = field(default_factory = dict)
    comments: dict[Id,Comment] = field(default_factory = dict)
    costumes: list[Costume] = field(default_factory = list)
    sounds: list[Sound] = field(default_factory = list)
    currentCostume: int = 1
    volume: int = 100
    
@dataclass
class Stage(Target):
    layerOrder: int = 0
    tempo: int = 120
    videoState: VideoState = VideoState.OFF
    videoTransparency: int = 50
    textToSpeechLanguage: str = 'English'
    
@dataclass
class Sprite(Target):
    layerOrder: int = 1
    visible: bool = True
    x: int = 0
    y: int = 0
    size: int = 100
    direction: int = 90
    draggable: bool = False
    rotationStyle: RotationStyle = RotationStyle.ALL_AROUND
    
@dataclass
class Project:
    targets: list[Sprite | Stage] = field(default_factory = list)
    monitors: list[Monitor] = field(default_factory = list)
    extensions: list[str] = field(default_factory = list)
    meta: Meta = field(default_factory = Meta)


def to_json(component: Any) -> Any:
    if isinstance(component,Asset) and not len(component.md5ext):
        component.md5ext = component.assetId + '.' + component.dataFormat 
    if isinstance(component,Variable):
        if component.cloud:
            return component
        else:
            return component.name,component.value
    elif isinstance(component,(Value,VariableValue,Input,Field,List)):
        return list(map(to_json,filter(lambda x: x is not None,component)))
    elif is_dataclass(component):
        return dict({k: to_json(getattr(component,k)) for k in component.__dataclass_fields__})
    elif isinstance(component,(tuple,list)):
        return list(map(to_json,component))
    elif isinstance(component,dict):
        return dict(zip(component.keys(),map(to_json,component.values())))
    elif isinstance(component,Enum):
        return component.value
    else:
        return component
    
def from_json(data: dict):
    return Project(**data)




from ..block import *
from ..extension import Extension

from enum import Enum

class SwitchBackdropToOptions(Enum):
    NEXT_BACKDROP = 'next backdrop'
    PREVIOUS_BACKDROP = 'previous backdrop'
    RANDOM_BACKDROP = 'random backdrop'

class EffectOptions(Enum):
    COLOR = 'COLOR'
    FISHEYE = 'FISHEYE'
    WHIRL = 'WHIRL'
    PIXELATE = 'PIXELATE'
    MOSAIC = 'MOSAIC'
    BRIGHTNESS = 'BRIGHTNESS'
    GHOST = 'GHOST'

class GoToFrontBackOptions(Enum):
    FRONT = 'front'
    BACK = 'back'

class GoForwardBackwardLayersOptions(Enum):
    FORWARD = 'forward'
    BACKWARD = 'backward'

class NumberNameOptions(Enum):
    NUMBER = 'number'
    NAME = 'name'


class BackdropNumberNameMonitor(Monitor):
    def __call__(self,builder: 'ScratchFileBuilder',option: str | NumberNameOptions,**kwargs) -> components.ListMonitor:
        monitor = super().__call__(builder,**kwargs)
        option = NumberNameOptions(option)
        monitor.id = monitor.id + '_' + option.value
        monitor.params = {'NUMBER_NAME': option}
        return monitor
    

class CostumeNumberNameMonitor(BackdropNumberNameMonitor,SpriteSpecificMonitor): pass


ARCHICAT_EXTENSION = Extension('looks')

ARCHICAT_EXTENSION.create_statements(
    Block('looks_switchbackdropto',BACKDROP=option_input(SwitchBackdropToOptions,SwitchBackdropToOptions.NEXT_BACKDROP,
                                                         Block('looks_backdrop',BACKDROP=option_field(SwitchBackdropToOptions)))
            ).attach_options(SwitchBackdropToOptions),
    access=BlockAccessModifier.ALL,
)

ARCHICAT_EXTENSION.create_statements(
    Block('looks_sayforsecs',MESSAGE=string_input,SECS=unsigned_float_input),
    Block('looks_say',MESSAGE=string_input),
    Block('looks_thinkforsecs',MESSAGE=string_input,SECS=unsigned_float_input),
    Block('looks_think',MESSAGE=string_input),
    Block('looks_switchcostumeto',COSTUME=costume_input),
    Block('looks_nextcostume'),
    Block('looks_changesizeby',SIZE=float_input),
    Block('looks_setsizeto',SIZE=unsigned_float_input),
    Block('looks_changeeffectby',EFFECT=option_field(EffectOptions),CHANGE=float_input).attach_options(EffectOptions),
    Block('looks_seteffect',EFFECT=option_field(EffectOptions),VALUE=unsigned_float_input).attach_options(EffectOptions),
    Block('looks_cleargraphicseffect'),
    Block('looks_show'),
    Block('looks_hide'),
    Block('looks_gotofrontback',FRONT_BACK=option_field(GoToFrontBackOptions)).attach_options(GoToFrontBackOptions),
    Block('looks_goforwardbackwardlayers',FORWARD_BACKWARD=option_field(GoForwardBackwardLayersOptions)
          ).attach_options(GoForwardBackwardLayersOptions),
    access=BlockAccessModifier.SPRITE,
)

ARCHICAT_EXTENSION.create_statements(
    Block('looks_switchbackdroptoandwait',BACKDROP=option_input(SwitchBackdropToOptions,SwitchBackdropToOptions.NEXT_BACKDROP,
                                                                Block('looks_backdrops',BACKDROP=option_field(SwitchBackdropToOptions)))
            ).attach_options(SwitchBackdropToOptions),
    access=BlockAccessModifier.STAGE,
)

ARCHICAT_EXTENSION.create_reporters(
    Block('looks_backdropnumbername',NUMBER_NAME=option_field(NumberNameOptions)).attach_options(NumberNameOptions),
    access=BlockAccessModifier.ALL,
)

ARCHICAT_EXTENSION.create_reporters(
    Block('looks_costumenumbername',NUMBER_NAME=option_field(NumberNameOptions)).attach_options(NumberNameOptions),
    Block('looks_size'),
    access=BlockAccessModifier.SPRITE,
)

ARCHICAT_EXTENSION.create_monitors(
    CostumeNumberNameMonitor('looks_costumenumbername').attach_options(NumberNameOptions),
    access=BlockAccessModifier.ALL
)

ARCHICAT_EXTENSION.create_monitors(
    BackdropNumberNameMonitor('looks_backdropnumbername').attach_options(NumberNameOptions),
    SpriteSpecificMonitor('looks_size'),
    access=BlockAccessModifier.SPRITE,
)